#!/usr/bin/perl -w

package RPM::Source::Convert::Mass;

use strict;
use warnings;
use Carp;
use Getopt::Long;
use File::Basename;
use RPM::Header;
use RPM qw/evrcmp/;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw/$DESTPATH $ORIGINPATH @ORIGINSKIP @DESTSKIP %DEST2ORIGINMAP/;
our ($DESTPATH, $ORIGINPATH);
my (@DESTPATH, @ORIGINPATH);

our @ORIGINSKIP=();
our @DESTSKIP=();
our %DEST2ORIGINMAP=();
my %ORIGINSKIP;
my %ORIGIN2DESTMAP;

our $OUTDIR="./OUT";

my $verbose;
our ($changelog, $release,$epoch, $rename, @hooks, $parent,$serial,$arg_tmpdir);
our ($skip_conversion_errors);
our $strategy = 'origin2dest';
our $origintoolsdir=dirname($0);
our $originimport="srpmconvert";

our %STRATEGYMAP=(
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destup' => \&destup,
    'newer' => \&newer,
    );

sub new {
    my $class = shift;
    my $self = {
	'-cliargs' => [],
	'-packages' => [],
	@_
    };
    bless $self, $class;
    my $result = GetOptions (
	"changelog=s"	=> \$changelog,
	"dest=s"	=> \$DESTPATH,
	"epoch=s"	=> \$epoch,
	"hook=s"	=> \@hooks,
	"origin=s"	=> \$ORIGINPATH,
	"out=s"		=> \$OUTDIR,
	"parent=s"	=> \$parent,
	"rename=s"	=> \$rename,
	"release=s"	=> \$release,
	"serial=s"	=> \$serial,
	"skip-errors"	=> \$skip_conversion_errors,
	"strategy=s"	=> \$strategy,
	"tmpdir=s"	=> \$arg_tmpdir,
	"v|verbose+"	=> \$verbose,
	@{$self->{'-cliargs'}}
	);

    @DESTPATH=&path_to_array($DESTPATH);
    @ORIGINPATH=&path_to_array($ORIGINPATH);
    
    if ((! scalar @DESTPATH) || (! scalar @ORIGINPATH)) {
	print "usage [ options ] [ list of rpms/names ]
options:
--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
--out	/path/to/output_dir (default is $OUTDIR)
--hook <code>	a subroutine hook. multiple hooks are allowed.
--tmpdir /path/	alternative tmpdir if default tmpdir is not big enough.
--changelog	'- changelog entry'
--epoch|serial|release|rename|parent <value>	force epoch/serial/release... value.
  usually it makes sense only for one package as the argument.
--skip-errors	ignore individual failures
--strategy 'strategy'	of mass operations (default is $strategy).
available strategies: ".join(',',keys(%STRATEGYMAP))."

usage examples:
upgrade all dest java packages using jpackage
originmass --dest /var/ftp/pub/Sisyphus/files/SRPMS --origin /var/ftp/pub/jpackage/1.7/generic/free/SRPMS
";
	exit (64);
    }

    %ORIGINSKIP=map {$_ => 1} @ORIGINSKIP;
    # TODO: implement auto-rename?
    %ORIGIN2DESTMAP=reverse (%DEST2ORIGINMAP);
    map {$ORIGINSKIP{$_}=1} map {$DEST2ORIGINMAP{$_} ? $DEST2ORIGINMAP{$_}: $_ } @DESTSKIP;

    push @{$self->{'-packages'}}, @ARGV;
    return $self;
}

sub run { 
    my ($self)=@_;
    mkdir $OUTDIR unless -e $OUTDIR;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    die "strategy $strategy not found" unless $strategy_allow;
    foreach my $destname (@{$self->{'-packages'}}) {
	# TODO:
	#if ($name=~/.src.rpm$/) {
	#$name=& MassCommon:: __get_rpm_name($rpm);
	#}
	my $originname=$destname;
	$originname=$DEST2ORIGINMAP{$destname} if $DEST2ORIGINMAP{$destname};
	next if $ORIGINSKIP{$originname};
	my $origin_rh=$self->find_origin_rpm($originname);
	#print STDERR 'found RPM '.$origin_rh->source_name." for $originname.\n";
	unless ($origin_rh) {
	    print STDERR "FATAL: foreign RPM $originname not found.\n";
	    next;
	}	
	my $dest_rh=$self->find_dest_rpm($destname);
	next unless &{$strategy_allow}($self,$originname, $destname,$origin_rh, $dest_rh);
	my ($originrpm, $destrpm);
	$originrpm=$origin_rh->source_name if $origin_rh;
	$destrpm=$dest_rh->source_name if $dest_rh;
	$self->__run_origin_import($originname,$originrpm,$destrpm) 
    }
}

sub __run_origin_import {
    my ($self,$originname, $originrpm, $destrpm)=@_;
    my @args;
    if ($verbose) {
	for (my $i=0; $i<$verbose-1; $i++) { # less verbose by 1; assumed verbose=1 is for mass only
	    push @args, '--verbose';
	}
    }
    push @args, '--outdir', $OUTDIR;
    push @args, '--changelog', $changelog if ($changelog);
    push @args, '--release', $release if ($release);
    push @args, '--rename', $rename if ($rename);
    push @args, '--epoch', $epoch if (defined $epoch);
    push @args, '--serial', $serial if (defined $serial);
    push @args, '--tmpdir', $arg_tmpdir if (defined $arg_tmpdir);
    if ($parent) {
	push @args, '--parent', $parent;
    } else {
	push @args, '--parent-rpm', $destrpm if ($destrpm);
    }
    if ($originrpm) {
	push @args, '--rpm', $originrpm;
    } else {
	print STDERR "FATAL: foreign RPM $originname not found.\n";
	return;
    }
    my $hook;
    $hook="$origintoolsdir/hooks/$originname.pl";
    push @args, '--hook', $hook if (-e $hook);
    foreach $hook (@hooks) {
	push @args, '--hook', $hook;
    }
    $self->origin_import_hook(\@args, $originname, $originrpm, $destrpm);

    print join(' ',$originimport,@args),"\n";
    if (system($originimport,@args)!=0) {
	print STDERR "ERROR: $originimport failed\n";
	die "exiting.\n" unless $skip_conversion_errors;
    }
    print "---*---\n";
}

sub origin_import_hook {}

sub find_origin_rpm {
    my ($self,$name)=@_;
    return $self->find_first_rpm($name,@ORIGINPATH);
}

sub find_dest_rpm {
    my ($self,$name)=@_;
    return $self->find_highest_rpm($name,@DESTPATH);
}

sub find_highest_rpm {
    my $self=shift;
    my $name=shift;
    my $cur_href;
    for my $path (@_) {
	for my $rpmfile (reverse(glob "$path/$name-*.src.rpm")) {
#	print STDERR "trying $rpmfile for $name\n";
	    my $href = new RPM::Header $rpmfile;
	    if ($name eq $href->{NAME}) {
		if (not $cur_href or $cur_href->cmpver($href)<0) {
		    $cur_href=$href;
		}
		# assume 1 rpm per name; false for jpp TODO:option!!!
		last;
	    }
	}
    }
    return $cur_href;
}

sub find_first_rpm {
    my $self=shift;
    my $name=shift;
    for my $path (@_) {
	for my $rpmfile (reverse(glob "$path/$name-*.src.rpm")) {
#	print STDERR "trying $rpm for $name\n";
	    my $href = new RPM::Header $rpmfile;
	    if ($name eq $href->{NAME}) {
		# assume 1 rpm per name; false for jpp TODO:option!!!
		return $href;
	    }
	}
    }
    return;
}

sub is_origin_up {
    my ($self, $origin_rh, $dest_rh)=@_;
    return 0 unless $origin_rh;
    return 1 unless $dest_rh;
    return $dest_rh->cmpver($origin_rh) <0;
}

#------------------------------------------

sub origin2dest {
    my ($self,$originname, $destname,$origin_rh, $dest_rh)=@_;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest_rh;
    return 1;
}

sub destup {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest_rh;
    return 1 if $dest_rh and $self->is_origin_up($origin_rh,$dest_rh);
    return 0;
}

sub newer {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "note: origin RPM $destname not found.\n" unless $origin_rh;
    return 1 if not $dest_rh or $self->is_origin_up($origin_rh,$dest_rh);
    return 0;
}

sub originskipdest {
    my ($self,$originname, $destname, $origin_rh, $dest_rh)=@_;
    print STDERR "found dest RPM $dest_rh for $originname.\n" if $dest_rh;
    return 0 if $dest_rh;# and $dest_rh!~/orphaned/;
    return 1;
}

#------------------------------------------
sub path_to_array {
    my $paths=shift;
    my @path;
    foreach (split(':',$paths)) {
	if ($_ eq '') {
	    $_ = '.';
	} elsif (/^~/) {
	    s/^~/$ENV{HOME}/;
	}
	push @path, $_ if -d $_;
    }
    return @path;
}


1;
